/*  Copyright (C) 2005,2006 Mokrushin I.V. aka McMCC mcmcc@mail.ru
    A simple parallel port programmer for I2C 8-pin serial EEPROMs (24XX devices).
    Based on original code by Alan Yates VK2ZAY alany@ay.com.au
    http://www.vk2zay.net/article.php/31

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
/* ------------------------------------------------------------------------- */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include "serp.h"
#include "i2c.h"

enum { EXPECT_ACK, SEND_ACK, SEND_NACK };

int i2cDelay = 0;
int IOBASE; /* typically 0x378 */

static void setupDelay() {
	/* the inb() and outb() calls include sufficient delay */
	int i;
	for(i = 0; i < i2cDelay; i++); /* needed for some devices */
}

static void pulldownSDA() {
	out_b((inb(IOBASE+2) | 0x02), IOBASE+2);
	setupDelay();
}

static void releaseSDA() {
	out_b((inb(IOBASE+2) & 0x04), IOBASE+2);
	setupDelay();
}

static void pulldownSCL() {
	out_b((inb(IOBASE+2) & 0xFB), IOBASE+2);
	setupDelay();
}

static void releaseSCL() {
	out_b((inb(IOBASE+2) | 0x04), IOBASE+2);
	setupDelay();
}

static int readSDA() {
	return ( 0 == ((inb(IOBASE + 2)) & 0x02) );
}

#if 0
static int readSCL() {
	return ( 4 == ((inb(IOBASE + 2)) & 0x04) );
}
#endif

/* asserts i2c start condition on the bus */
static void i2cStart() {
	/* first we ensure both lines high
	  (this is also a stop condition if the data is currently low, a useful side effect) */
	releaseSCL();
	releaseSDA();

	/* then we drive the data line low to signal the start condition */
	pulldownSDA();

	/* now we can pull down the clock ready to send data */
	pulldownSCL();
}

/* asserts i2c stop condition on the bus */
static void i2cStop() {
	/* first we ensure clock is high and data is low */
	pulldownSDA();
	releaseSCL();

	/* then we drive data high to signal the stop condition */
	releaseSDA();
}

/*
 * Sends byte over the i2c bus
 * Will return 0 on sucessful (acked) transfer)
 * Non-zero otherwise
 */
static int i2cSendByte(int byte) {
	int i, ret;

	/* the clock should be currently low */
	
	/* send byte MSB first */
	for(i = 7; i > -1; i--) {
		(byte & (1 << i))?(releaseSDA()):(pulldownSDA());

		/* clock */
		releaseSCL();
		pulldownSCL();
	}

	/* get ACK - data has to be released so slave can pull it down */
	releaseSDA();
	releaseSCL();
	ret = readSDA();
	pulldownSCL();

	return ret;
}

/*
 * Reads a byte off the i2c bus
 * Will return byte value on sucessful (acked) transfer
 * Negative if ack was requested and failed
 */
static int i2cReadByte(int ackControl) {
	int i, val = 0, ack;

	/* read byte MSB first */
	for(i = 7; i > -1; i--) {
		releaseSDA();
		releaseSCL();
		val |= (readSDA() << i);
		pulldownSCL();
		pulldownSDA();
	}

	switch(ackControl) {
		case EXPECT_ACK:
			releaseSCL();
			ack = readSDA();
			pulldownSCL();
			if(ack) return -1;
			break;
		case SEND_ACK:
			pulldownSDA();
			releaseSCL();
			pulldownSCL();
			releaseSDA();
			break;
		case SEND_NACK:
			releaseSDA();
			releaseSCL();
			pulldownSCL();
			pulldownSDA();
			break;
		default:
			fprintf(stderr, "error: unknown ack control mode!\n");
			i2cStop();
			exit(-1);
	}

	return val;
}

/*
 * set busy-wait delay between i2c transitions
 */
void i2cSetDelay(int delay) {
	i2cDelay = delay;
}

/*
int i2cInit(int baseAddress) {
	IOBASE = baseAddress;
*/

int i2cInit() {
	IOBASE = LPT; /* typically 0x378 */
	/* power on  */
	out_b(0x10, IOBASE);					
	out_b(((inb(IOBASE + 2) & 0xF0) | 0x6), IOBASE+ 2);
	usleep(5);
	return 0;
}

int i2cCleanUp() {
	/* power off */
	out_b(0, IOBASE);
	out_b(((inb(IOBASE + 2) & 0xF0) | 0x9), IOBASE + 2);
	return 0;
}

int i2c_deviceSize(int device) {
	switch(device) {
		case 1:		return 0x80;
		case 2:		return 0x100;
		case 4:		return 0x200;
		case 8:		return 0x400;
		case 16:	return 0x800;
		case 32:	return 0x1000;
		case 64:	return 0x2000;
		case 128:	return 0x4000;
		case 256:	return 0x8000;
		default:
				fprintf(stderr, "unknown device 24C%d\n", device);
				return -1;
	}
}

int i2c_readRandom(int device, int address,  char *buf, int len) {
	int i, ret, controlByte = 0;
	i2cStart();

	/* set read point by doing a partial write */
	switch(device) {
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
			controlByte = 0xA0 | ((address & 0x700) >> 7);
			if(i2cSendByte(controlByte)) {
				fprintf(stderr, "read[0x%.3X]: control byte 1 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			if(i2cSendByte(address & 0xFF)) {
				fprintf(stderr, "read[0x%.3X]: addressing byte 1 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			break;
		case 32:
		case 64:
		case 128:
		case 256:
			controlByte = 0xA0;
			if(i2cSendByte(controlByte)) {
				fprintf(stderr, "read[0x%.3X]: control byte 1 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			if(i2cSendByte((address >> 8) & 0xFF)) {
				fprintf(stderr, "read[0x%.3X]: addressing high byte 1 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			if(i2cSendByte(address & 0xFF)) {
				fprintf(stderr, "read[0x%.3X]: addressing low byte 1 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			break;
		default:
			fprintf(stderr, "read[0x%.3X]: unknown device %d\n", address, device);
			break;
	}

	/*
	 * read and write operations require control byte 4 MSB to be 1010
	 * read sets LSB to 1, write 0
	 * the middle three bits are the (256 byte) block address
	 */
	i2cStart();
	switch(device) {
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
			controlByte = 0xA1 | ((address & 0x700) >> 7);
			if(i2cSendByte(controlByte)) {
				fprintf(stderr, "read[0x%.3X]: control byte 2 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			break;
		case 32:
		case 64:
		case 128:
		case 256:
			controlByte = 0xA1;
			if(i2cSendByte(controlByte)) {
				fprintf(stderr, "read[0x%.3X]: control byte 2 was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			break;
		default:
			/* not reached */
			break;
	}
 	for(i = 0; i < len; i++) {
		ret = i2cReadByte(((i == len - 1)?(SEND_NACK):(SEND_ACK)));
		if(ret < 0) {
			fprintf(stderr, "read[0x%.3X]: byte read failed\n", address + i);
			i2cStop();
			return -1;
		}
		buf[i] = ret & 0xFF;
	}

	i2cStop();
	return 0;
}

/*
 * writes the byte value to the specified address
 * returns true on error, zero other wise
 */
int i2c_writeByte(int device, int address, int value) {
	int controlByte = 0, busy;
	i2cStart();

	/* set control and address */
	switch(device) {
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
			controlByte = 0xA0 | ((address & 0x700) >> 7);
			if(i2cSendByte(controlByte)) {
				fprintf(stderr, "write[0x%.3X]: control byte was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			if(i2cSendByte(address & 0xFF)) {
				fprintf(stderr, "write[0x%.3X]: addressing byte was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			break;
		case 32:
		case 64:
		case 128:
		case 256:
			controlByte = 0xA0;
			if(i2cSendByte(controlByte)) {
				fprintf(stderr, "write[0x%.3X]: control byte was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			if(i2cSendByte((address >> 8) & 0xFF)) {
				fprintf(stderr, "write[0x%.3X]: addressing high byte was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			if(i2cSendByte(address & 0xFF)) {
				fprintf(stderr, "write[0x%.3X]: addressing low byte was not ACKed\n", address);
				i2cStop();
				return -1;
			}
			break;
		default:
			fprintf(stderr, "read[0x%.3X]: unknown device %d\n", address, device);
			break;
	}
	    
	/* send data byte */
	if(i2cSendByte(value & 0xFF)) {
		fprintf(stderr, "write[0x%.3X]: data byte was not ACKed\n", address);
		i2cStop();
		return -1;
	}

	/* write begins after STOP signal */
	i2cStop();
	
	/* no we poll for acknowledgement on completion of write */
	for(busy = 0; busy < 1000; busy ++) {
		i2cStart();
		if(!i2cSendByte(controlByte)) {
			if(busy < 1) {
				fprintf(stderr, "write[0x%.3X]: write ACKed immediately, something is probably wrong!\n", address);
			}
			i2cStop();
			return 0;
		}
	}

	fprintf(stderr, "write[0x%.3X]: waited %d cycles for ACK, failed!\n", address, busy);
	i2cStop();
	return -1;
}

int i2c_writeRandom(int device, int address,  char *buf, int len) {
	int i;

	for(i = 0; i < len; i++) {
		if(i2c_writeByte(device, address + i, buf[i])) return -1;
	}

	return 0;
}
