/*  Copyright (C) 2005,2006 Mokrushin I.V. aka McMCC mcmcc@mail.ru
    A simple parallel port programmer for Microwire 8-pin serial EEPROMs 
    (93XX devices). Support organization 8bit and 16bit(8bit emulation).

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
/* ------------------------------------------------------------------------- */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include "serp.h"
#include "microwire.h"

int PORT;

static unsigned char VCC  = 0x10;
static unsigned char ORG  = 0x8; /* organization 0 = 8 bit and 1 = 16 bit */
static unsigned char CLK  = 0x2; 
static unsigned char DO   = 0x10;
static unsigned char DI   = 0x4;
static unsigned char CSEL = 0x1;

static unsigned char data;

static void delay_ms(int n)
{
#ifndef _WINDOWS
	int i;
	for(i = 0; i < (n * 100000); i++);
#else
	usleep(n);
#endif
}

static void data_1()
{
        data = data | DI;
        out_b(data, PORT);
}

static void data_0()
{
        data = data & (~DI);
        out_b(data, PORT);
}

static void org_1()
{
        data = data | ORG; /* 16bit */
        out_b(data, PORT);
}

static void org_0()
{
        data = data & (~ORG); /* 8bit */
        out_b(data, PORT);
}

static void csel_1()
{
        data = data | CSEL;
        out_b(data, PORT);
}

static void csel_0()
{
        data = data & (~CSEL);
        out_b(data, PORT);
}

static void clock_1()
{
        data = data | CLK;
        out_b(data, PORT);
}

static void clock_0()
{
        data = data & (~CLK);
        out_b(data, PORT);
}

static void init_port()
{
	PORT = LPT;
        data = 0x16;
        out_b(data, PORT);
}

static void power_on()
{
        data = data | VCC;
        out_b(data, PORT);
}

static void power_off()
{
	org_0();
        data = data & (~VCC);
        out_b(data, PORT);
}

static unsigned int get_data()
{
        unsigned int b;
        b = inb(PORT+1);
        return ((b & DO) == DO);
}

static int addr_nbits(int size)
{
	int i;
	i = 0;
    
	while(size > 0)
	{
    	    size = size >> 1;
	    if(size)
    		i++;
	}
	if(i < 6) /* Fix for 93C06 */
	    i = 6;
	return i;
}

static int convert_size(int eeprom_size, int org)
{
	int k;
	org_0();
	delay_ms(1);
	k = 1;

	if(org)
	{
	    org_1();
	    delay_ms(1);
	    k = 2;
	}    
	eeprom_size = eeprom_size / k;
	return eeprom_size;
}

static void send_to_di(unsigned char val, int nbit)
{
	int b, i;
	i = 0;

	while(i < nbit)
	{
		b = val & (1<<(nbit-1));
		clock_0();
		if(b) 	
		    data_1();
		else
		    data_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		i++;
		val = (val & ((1<<(nbit-1))-1)) * 2;
	}
}

static unsigned char get_from_do()
{
	unsigned char val;
	int b, i;
	i = 0;
	val = 0;
	while( i < 8)
	{
		val = val * 2;
		clock_0();
  		delay_ms(1);
		b = get_data();
		clock_1();
		delay_ms(1);
		i++;
		val += b;
	}
	return val;
}

static void enable_write_3wire(int num_bit)
{
	csel_0();
	clock_0();
	delay_ms(1);
	data_1();
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	send_to_di(3, 4);
	send_to_di(0, num_bit-2);

	data_0();
	delay_ms(1);
	csel_0();
	delay_ms(1);
}

static void disable_write_3wire(int num_bit)
{
	csel_1();
	delay_ms(1);
	clock_0();
	delay_ms(1);
	data_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);
	send_to_di(0, 4);
	send_to_di(0, num_bit-2);
	csel_0();
	delay_ms(1);
}

static void chip_busy(void)
{
	fprintf(stderr, "Error: Always BUSY! Communication problem...The broken chip?\n");
	power_off();
}

void Erase_EEPROM_3wire(int size_eeprom, int org)
{
	int i, num_bit;

	init_port();
	power_on();

	size_eeprom = convert_size(size_eeprom, org);
	num_bit = addr_nbits(size_eeprom);
	        
	enable_write_3wire(num_bit);
	csel_0();
	clock_0();
	delay_ms(1);
	data_1();
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	send_to_di(2, 4);
	send_to_di(0, num_bit-2);
	clock_0();
	data_0();
	delay_ms(1);
	csel_0();
	delay_ms(1);
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);
	i = 0;
	while(!get_data() && i < 10000)
	{
		clock_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		i++;
	}
	
	if(i == 10000)
	{
		chip_busy();
		return;
	}

	csel_0();
	delay_ms(1);
	clock_0();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	disable_write_3wire(num_bit);
	power_off();

}

unsigned char* Read_EEPROM_3wire(unsigned char *buffer, int size_eeprom, int org)
{
	int address, num_bit, l;

	init_port();
	power_on();

	size_eeprom = convert_size(size_eeprom, org);
	num_bit = addr_nbits(size_eeprom);

	address = 0;

	for(l = 0; l < size_eeprom; l++)
	{
		csel_0();
		clock_0();
		delay_ms(1);
		data_1();
		csel_1();
		delay_ms(1);
		clock_1();
		delay_ms(1);

		send_to_di(2, 2);
		send_to_di(l, num_bit);

		data_0();
		clock_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		buffer[address] = get_from_do();
		if(org)
		{
		    address++;
		    buffer[address] = get_from_do();
		}
		csel_0();
		delay_ms(1);
		clock_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		address++;

	}
	power_off();
	return buffer;
}

unsigned char* Read_byte_EEPROM_3wire(unsigned char *buffer, int address, int size_eeprom, int org)
{
	int num_bit;

	init_port();
	power_on();

	address = address >> org;
	size_eeprom = convert_size(size_eeprom, org);
	num_bit = addr_nbits(size_eeprom);


	csel_0();
	clock_0();
	delay_ms(1);
	data_1();
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	send_to_di(2, 2);
	send_to_di(address, num_bit);

	data_0();
	clock_0();
	delay_ms(1);
	clock_1();
	delay_ms(1);
	buffer[0] = get_from_do();
	if(org)
	{
	    buffer[1] = get_from_do();
	}
	csel_0();
	delay_ms(1);
	clock_0();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	power_off();
	return buffer;
}

int Write_byte_EEPROM_3wire(unsigned char val, int address, int size_eeprom, int org)
{
	int num_bit, seek = 0;
	unsigned char *buf;
	int i = 0;
		
	buf = malloc(2);
	buf = Read_byte_EEPROM_3wire(buf, address, size_eeprom, org);

	init_port();
	power_on();

	seek = address & org;
	address = address >> org;
	buf[seek] = val;
	        
	size_eeprom = convert_size(size_eeprom, org);
	num_bit = addr_nbits(size_eeprom);
	

	enable_write_3wire(num_bit);

	csel_0();
	clock_0();
	delay_ms(1);
	data_1();
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);
	send_to_di(1, 2);
	send_to_di(address, num_bit);
	send_to_di(buf[0], 8);
	if(org)
	{
	    send_to_di(buf[1], 8);
	}
	clock_0();
	data_0();
	delay_ms(1);
	csel_0();
	delay_ms(1);
	csel_1();
	delay_ms(1);
	clock_1();

	delay_ms(1);
	while (!get_data() && i<10000)
	{
		clock_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		i++;
	}
	if (i==10000)
	{
		chip_busy();
		if(buf) free(buf);
		return -1;
	}

	csel_0();
	delay_ms(1);
	clock_0();
	delay_ms(1);
	clock_1();
	delay_ms(1);
	disable_write_3wire(num_bit);
	power_off();
	if(buf) free(buf);
	return 0;
}

int Write_EEPROM_3wire(unsigned char *buffer, int size_eeprom, int org)
{
	int i, l, address, num_bit;

	init_port();
	power_on();

	size_eeprom = convert_size(size_eeprom, org);
	num_bit = addr_nbits(size_eeprom);
	        
	enable_write_3wire(num_bit);
	address = 0;

	for(l = 0; l < size_eeprom; l++)
	{
		csel_0();
		clock_0();
		delay_ms(1);
		data_1();
		csel_1();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		send_to_di(1, 2);
		send_to_di(l, num_bit);
		send_to_di(buffer[address], 8);
		if(org)
		{
		    address++;
		    send_to_di(buffer[address], 8);
		}
		clock_0();
		data_0();
		delay_ms(1);
		csel_0();
		delay_ms(1);
		csel_1();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		i = 0;
		while (!get_data() && i<10000)
		{
			clock_0();
			delay_ms(1);
			clock_1();
			delay_ms(1);
			i++;
		}
		if (i==10000)
		{
			chip_busy();
			return -1;
		}

		csel_0();
		delay_ms(1);
		clock_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);

		address++;

	}
	disable_write_3wire(num_bit);
	power_off();
	return 0;
}

int Write_fill_EEPROM_3wire(unsigned char val, int size_eeprom, int org)
{
	int i, num_bit;

	init_port();
	power_on();

	size_eeprom = convert_size(size_eeprom, org);
	num_bit = addr_nbits(size_eeprom);
	        
	enable_write_3wire(num_bit);
	csel_0();
	clock_0();
	delay_ms(1);
	data_1();
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	send_to_di(1, 4);
	send_to_di(0, num_bit-2);
	send_to_di(val, 8);
	if(org)
	    send_to_di(val, 8);
	clock_0();
	data_0();
	delay_ms(1);
	csel_0();
	delay_ms(1);
	csel_1();
	delay_ms(1);
	clock_1();
	delay_ms(1);
	i = 0;
	while(!get_data() && i < 10000)
	{
		clock_0();
		delay_ms(1);
		clock_1();
		delay_ms(1);
		i++;
	}
	
	if(i == 10000)
	{
		chip_busy();
		return -1;
	}

	csel_0();
	delay_ms(1);
	clock_0();
	delay_ms(1);
	clock_1();
	delay_ms(1);

	disable_write_3wire(num_bit);
	power_off();
	return 0;
}

int deviceSize_3wire(int device) 
{
        switch(device) {
		case  6:	 return 32;
		case 16:	 return 64;
                case 46:         return 128;
                case 56:         return 256;
                case 66:         return 512;
                case 76:         return 1024;
                case 86:         return 2048;
                case 96:         return 4096;
                default:
                                fprintf(stderr, "unknown device 93C%d\n", device);
                                return -1;
        }
}
