/*  Copyright (C) 2005,2006 Mokrushin I.V. aka McMCC mcmcc@mail.ru
    
    SERP - Serial EEPROM 24CXX/93CXX programmer on parallel port.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
/* ------------------------------------------------------------------------- */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifndef __FreeBSD__
#if defined(_WINDOWS) && !defined(_IOPERM)
#ifndef NO_PORT_TALK
#include "pt_ioctl.h"
#endif
#else
#ifndef WATCOM
#include <sys/io.h>
#endif
#endif
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <time.h>
#include <ctype.h>

#include "serp.h"
#include "i2c.h"
#include "microwire.h"

#ifdef _WINDOWS
static char *file_open_dlg()
{
	OPENFILENAME ofn;
	char szFile[256];
	
	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.lpstrFile = szFile;
	
	ofn.lpstrFile[0] = '\0';
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "EEPROM Files\0*.bin\0All\0*.*\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
	
	if (GetOpenFileName(&ofn)==TRUE)
	    return ofn.lpstrFile;
	return NULL;
}

static char *file_save_dlg()
{
	OPENFILENAME ofn;
	char szFile[256];
	
	ZeroMemory(&ofn, sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.lpstrFile = szFile;
	
	ofn.lpstrFile[0] = '\0';
	ofn.lpstrDefExt = ".bin\0";
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "EEPROM Files\0*.bin\0All\0*.*\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
	
	if (GetSaveFileName(&ofn)==TRUE)
	    return ofn.lpstrFile;
	return NULL;
}
#endif

static void hexdump(char *buf, int len, int offset) 
{
	int i, j;

	for(i = 0; i < len; i+= 16) {
		printf("0x%.4X: ", i + offset);
		for(j = 0; (j < 16) && (i+j < len); j++)
			printf("%.2X ", buf[i+j] & 0xFF);
		for( ; j < 16; j++)
			printf("   ");
		printf("    ");
		for(j = 0; (j < 16) && (i+j < len); j++)
			if(isprint(buf[i+j]))
				printf("%c", buf[i+j]);
			else
				printf(".");
		printf("\n");
	}
}

static int verify(int type, int device, char *buf, char val, int bit16)
{
	int size_eeprom, address;
	char *buff;

	if(type != 24 && type != 93)
	{
	    fprintf(stderr, "unknown chip EEPROM %d\n", type);
	    return -1;
	}
	
	if(type == 24)
	    size_eeprom = i2c_deviceSize(device);
	else
	    size_eeprom = deviceSize_3wire(device);

	buff = malloc(size_eeprom);
	memset(buff, 0, size_eeprom);

	if(type == 24)
	{
	    if(i2c_readRandom(device, 0, buff, size_eeprom)) {
		fprintf(stderr, "read problem from EEPROM!\n");
		if(buff) free(buff);
		return -1;
	    }
	} else {
	    buff = Read_EEPROM_3wire(buff, size_eeprom, bit16);
	    if(buff == NULL) {
		fprintf(stderr, "read problem from EEPROM!\n");
		if(buff) free(buff);
		return -1;
	    }
	}
	
	if(buf != NULL)
	{
	    for(address = 0; address < size_eeprom; address++)
	    {
		if(buf[address] != buff[address])
		{
		    fprintf(stderr, "verify EEPROM...BAD!\n");
		    if(buff) free(buff);
		    return -1;
		}
	    }
	} else {
	    for(address = 0; address < size_eeprom; address++)
	    {
		if(val != buff[address])
		{
		    fprintf(stderr, "verify EEPROM...BAD!\n");
		    if(buff) free(buff);
		    return -1;
		}
	    }
	}
	if(buff) free(buff);
	fprintf(stderr, "verify EEPROM...OK!\n");
	return 0;
}

static void dump(int type, int device, int start, int len, int bit16) 
{
	char *buf;
	int size_eeprom;
	
	if(type != 24 && type != 93)
	{
	    fprintf(stderr, "unknown chip EEPROM %d\n", type);
	    return;
	}
	
	if(type == 24)
	    size_eeprom = i2c_deviceSize(device);
	else
	    size_eeprom = deviceSize_3wire(device);
	    
	if(start > (size_eeprom-1))
	{
	    fprintf(stderr, "Address 0x%.4X is not valid, as the chip has max length = %d bytes\n", start, size_eeprom);
	    return;
	}

	if(len > size_eeprom)	
	{
	    fprintf(stderr, "Length %d is not valid, as the chip has max length = %d bytes\n", len, size_eeprom);
	    return;
	}

	if(len <= 0)	
	{
	    fprintf(stderr, "Length %d is not valid\n", len);
	    return;
	}

	buf = malloc(size_eeprom);
	memset(buf, 0, size_eeprom);

	if(type == 24)
	{
	    if(!i2c_readRandom(device, start, buf, len)) {
		    hexdump(buf, len, start);
	    }
	} else {
	    buf = Read_EEPROM_3wire(buf, size_eeprom, bit16);
	    hexdump(buf+start, len, start);
	}
	if(buf) free(buf);
}

static int dumpToFile(int type, int device, char *filename, int bit16) 
{
	FILE *fp; 
	int ret, len = 0;
	char *buf;

	if(type != 24 && type != 93)
	{
	    fprintf(stderr, "unknown chip EEPROM %d\n", type);
	    return -1;
	}

	if(type == 24)
	    len = i2c_deviceSize(device);
	else
	    len = deviceSize_3wire(device);
	
	buf = malloc(len);
	memset(buf, 0, len);
	
	if(type == 24)
	{
	    if(i2c_readRandom(device, 0, buf, len))
	    {
	        if(buf) free(buf);
    	        return -1;
	    }
	} else {
	    buf = Read_EEPROM_3wire(buf, len, bit16);
	}

	fp = fopen(filename, "wb");
	if(!fp) {
		perror("fopen()");
		if(buf) free(buf);
		return -1;
	}

	ret = fwrite(buf, 1, len, fp);
	if(ret != len) {
		perror("fwrite()");
		if(buf) free(buf);
		fclose(fp);
		return -1;
	}
	fclose(fp);

	fprintf(stderr, "copied %d bytes\n", ret);
	if(buf) free(buf);
	return 0;
}

static int programFromFile(int type, int device, char *filename, int bit16) 
{
	FILE *fp; 
	int ret, len = 0;
	char *buf;

	if(type != 24 && type != 93)
	{
	    fprintf(stderr, "unknown chip EEPROM %d\n", type);
	    return -1;
	}

	if(type == 24)
	    len = i2c_deviceSize(device);
	else
	    len = deviceSize_3wire(device);
	
	buf = malloc(len);
	memset(buf, 0, len);

	fp = fopen(filename, "rb");
	if(!fp) {
		perror("fopen()");
		if(buf) free(buf);
		return -1;
	}

	ret = fread(buf, 1, len, fp);
	if(ret < 1) {
		perror("fread()");
		if(buf) free(buf);
		fclose(fp);
		return -1;
	}
	fclose(fp);

	if(type == 24)
	{
	    if(i2c_writeRandom(device, 0, buf, ret))
	    {
		if(buf) free(buf);
	        return -1;
	    }
	} else {
	    if(Write_EEPROM_3wire(buf, len, bit16))
	    {
		if(buf) free(buf);
	        return -1;
	    }
	}
	fprintf(stderr, "programmed %d bytes\n", ret);
	verify(type, device, buf, 0, bit16);
	if(buf) free(buf);
	return 0;
}

static void fill(int type, int device, char b, int bit16) 
{
	int i, len;

	if(type != 24 && type != 93)
	{
	    fprintf(stderr, "unknown chip EEPROM %d\n", type);
	    return;
	}
	
	if(type == 24)
	{
	    len = i2c_deviceSize(device);
	    for(i = 0; i < len; i++) {
		if(i2c_writeByte(device, i, b)) return;
	    }
	} else {
	    len = deviceSize_3wire(device);
	    if(Write_fill_EEPROM_3wire(b, len, bit16)) return;	
	}
	fprintf(stderr, "cleared %d bytes\n", len);
	verify(type, device, NULL, b, bit16);
}

static void edit(int type, int device, int address, int b, int bit16) 
{
	int len;

	if(type != 24 && type != 93)
	{
	    fprintf(stderr, "unknown chip EEPROM %d\n", type);
	    return;
	}
	
	if(type == 24)
	    len = i2c_deviceSize(device);
	else
	    len = deviceSize_3wire(device);

	if(address > (len-1))
	{
	    fprintf(stderr, "Address 0x%.4X is not valid, as the chip has max length = %d bytes\n", address, len);
	    return;
	}

	if(type == 24)
	{
	    if(i2c_writeByte(device, address, b)) return;
	} else {
	    if(Write_byte_EEPROM_3wire(b, address, len, bit16)) return;
	}	
	fprintf(stderr, "Write byte 0x%.2X to address 0x%.4X\n", b, address);
}

static void about() {
	fprintf(stderr, "\nSERP(Serial EEPROM Programmer) " VERSION "\n");
	fprintf(stderr, "Copyright (C) 2005,2006 Mokrushin I.V. aka McMCC <mcmcc@mail.ru>\n");
	fprintf(stderr, "Supports i2c: 24C01,24C02,24C04,24C08,24C16,24C32,24C64,24C128,24C256.\n");
	fprintf(stderr, "Supports microwire: 93C06,93C46,93C56,93C66,93C76,93C86,93C96\n");
	fprintf(stderr, "Platform: ");
#ifndef _WINDOWS
#ifdef __FreeBSD__
	fprintf(stderr, "FreeBSD\n");
#else
	fprintf(stderr, "Linux\n");
#endif
#else /* Windows */
#ifdef NO_PORT_TALK
#ifdef _IOPERM
	fprintf(stderr, "Win32 with GNU Cygwin IoPerm driver\n");
#else
	fprintf(stderr, "Win32 with InpOut32 driver\n");
#endif
#else
	fprintf(stderr, "Win32 with PortTalk driver\n");
#endif
#endif
#if defined(BUILD) && defined(CPU)
	fprintf(stderr, "Build: " BUILD "." CPU "\n");
#endif
	fprintf(stderr, "\n");
}

static void usage() {
	fprintf(stderr, "Commands:\n");
	fprintf(stderr, "\tq\t\tExit from programm\n");
	fprintf(stderr, "\t?\t\tThis usage message\n");
	fprintf(stderr, "\ti\t\tSelect i2c EEPROM mode for 24CXX\n");
	fprintf(stderr, "\tm\t\tSelect microwire EEPROM mode for 93CXX\n");
}

static void usage_24xx() {
	fprintf(stderr, "Commands:\n");
	fprintf(stderr, "\tq\t\tQuit\n");
	fprintf(stderr, "\t?\t\tThis usage message\n");
	fprintf(stderr, "\td\t\tDump EEPROM, as hex\n");
	fprintf(stderr, "\tD <start> <len>\tDump EEPROM, as hex\n");
	fprintf(stderr, "\te <addr> <val>\tEdit EEPROM, write value to address\n");
	fprintf(stderr, "\tf <value>\tErase EEPROM, filling with value\n");
#ifdef _WINDOWS
	fprintf(stderr, "\tr\t\tDump EEPROM into raw binary file\n");
	fprintf(stderr, "\tw\t\tProgram EEPROM from raw binary file\n");
#else
	fprintf(stderr, "\tr <filename>\tDump EEPROM into raw binary file\n");
	fprintf(stderr, "\tw <filename>\tProgram EEPROM from raw binary file\n");
#endif
	fprintf(stderr, "\tt <type>\tSet EEPROM device type (as 24C<type>), default 1\n");
	fprintf(stderr, "\tp <delay>\tSet delay between I2C transitions (for slower devices)\n");
}

static void usage_93xx() {
	fprintf(stderr, "Commands:\n");
	fprintf(stderr, "\tq\t\tQuit\n");
	fprintf(stderr, "\tc\t\tFull erase EEPROM\n");
	fprintf(stderr, "\t?\t\tThis usage message\n");
	fprintf(stderr, "\td\t\tDump EEPROM, as hex\n");
	fprintf(stderr, "\tD <start> <len>\tDump EEPROM, as hex\n");
	fprintf(stderr, "\te <addr> <val>\tEdit EEPROM, write value to address\n");
	fprintf(stderr, "\tf <value>\tErase EEPROM, filling with value\n");
#ifdef _WINDOWS
	fprintf(stderr, "\tr\t\tDump EEPROM into raw binary file\n");
	fprintf(stderr, "\tw\t\tProgram EEPROM from raw binary file\n");
#else
	fprintf(stderr, "\tr <filename>\tDump EEPROM into raw binary file\n");
	fprintf(stderr, "\tw <filename>\tProgram EEPROM from raw binary file\n");
#endif
	fprintf(stderr, "\tt <type>\tSet EEPROM device type (as 93C<type>), default 46\n");
	fprintf(stderr, "\tb <bit>\t\tSet bit organization 8 or 16, default 16\n");
}

static void main_24xx() {
	int a, b, device = 1, run = 1, st = 0;
	char c, buf[0xFF];
#ifdef _WINDOWS
	char *filen;
#endif
	int type = 24;

	while(run) {
		if(scanf("%c", &c) < 1) break;
		switch(c) {
			case 't':
				scanf("%i", &a);
				b = i2c_deviceSize(a);
				if(b < 0) continue;
				device = a;
				break;
			case 'p':
				scanf("%i", &a);
				if(a < 0) a = 0;
				i2cSetDelay(a);
				fprintf(stderr, "delay set to %d loops\n", a);
				break;
			case 'd':
				if(i2cInit()) break;
				dump(type, device, 0, i2c_deviceSize(device), 0);
				i2cCleanUp();
				break;
			case 'D':
				scanf("%i", &a);
				scanf("%i", &b);
				if(i2cInit()) break;
				dump(type, device, a, b, 0);
				i2cCleanUp();
				break;
			case 'e':
			case 'E':
				scanf("%i", &a);
				scanf("%i", &b);
				if(i2cInit()) break;
				edit(type, device, a, 0xFF & b, 0);
				i2cCleanUp();
				break;
			case 'f':
			case 'F':
				scanf("%i", &b);
				if(i2cInit()) break;
				fill(type, device, 0xFF & b, 0);
				i2cCleanUp();
				break;
			case 'r':
			case 'R':
#ifdef _WINDOWS
				filen = file_save_dlg();
				if(filen)
					strncpy(buf, filen, 256);
				else
					break;
#else
				scanf("%s", buf);
#endif
				if(i2cInit()) break;
				dumpToFile(type, device, buf, 0);
				i2cCleanUp();
				break;
			case 'w':
			case 'W':
#ifdef _WINDOWS
				filen = file_open_dlg();
				if(filen)
					strncpy(buf, filen, 256);
				else
					break;
#else
				scanf("%s", buf);
#endif
				if(i2cInit()) break;
				programFromFile(type, device, buf, 0);
				i2cCleanUp();
				break;
			case 'q':
			case 'Q':
				run = 0;
				break;
			case '\n':
				st = 1;
				break;
			case 'h':
			case 'H':
			case '?':
				usage_24xx();
				break;
		}
		if(run && st)
		{
		    fprintf(stderr, "24C%.2d> ", device);
		    st = 0;
		}
	}

	return;
}

static void main_93xx() {
	int a, b, s, device = 46, run = 1, st = 0;
	char c, buf[0xFF];
#ifdef _WINDOWS
	char *filen;
#endif
	int type = 93;
	int bit_16 = 1;

	while(run) {
		if(scanf("%c", &c) < 1) break;
		switch(c) {
			case 't':
				scanf("%i", &a);
				b = deviceSize_3wire(a);
				if(b < 0) continue;
				device = a;
				break;
			case 'b':
				scanf("%i", &a);
				if(a != 8 && a != 16) {
				    fprintf(stderr, "set bit organization %d invalid\n", a);
				    fprintf(stderr, "set only 8 or 16 bit!\n");
				    break;
				}
				if(a == 8)
				    bit_16 = 0;
				else
				    bit_16 = 1;
				break;
			case 'd':
				dump(type, device, 0, deviceSize_3wire(device), bit_16);
				break;
			case 'D':
				scanf("%i", &a);
				scanf("%i", &b);
				dump(type, device, a, b, bit_16);
				break;
			case 'e':
			case 'E':
				scanf("%i", &a);
				scanf("%i", &b);
				edit(type, device, a, 0xFF & b, bit_16);
				break;
			case 'f':
			case 'F':
				scanf("%i", &b);
				fill(type, device, 0xFF & b, bit_16);
				break;
			case 'r':
			case 'R':
#ifdef _WINDOWS
				filen = file_save_dlg();
				if(filen)
					strncpy(buf, filen, 256);
				else
					break;
#else
				scanf("%s", buf);
#endif
				dumpToFile(type, device, buf, bit_16);
				break;
			case 'c':
			case 'C':
				s = deviceSize_3wire(device);
				Erase_EEPROM_3wire(s, bit_16);
				verify(type, device, NULL, 0xFF, bit_16);
				break;
			case 'w':
			case 'W':
#ifdef _WINDOWS
				filen = file_open_dlg();
				if(filen)
					strncpy(buf, filen, 256);
				else
					break;
#else
				scanf("%s", buf);
#endif
				programFromFile(type, device, buf, bit_16);
				break;
			case 'q':
			case 'Q':
				run = 0;
				break;
			case '\n':
				st = 1;
				break;
			case 'h':
			case 'H':
			case '?':
				usage_93xx();
				break;
		}
		if(run && st)
		{
		    fprintf(stderr, "93C%.2d %dbit> ", device, bit_16?16:8);
		    st = 0;
		}
	}

	return;
}

int main(int argc, char **argv) {
	int run = 1, st = 0;
	char c;

	about();
#if !defined(_WINDOWS) || defined(_IOPERM)
#ifdef __FreeBSD__
        if(i386_set_ioperm(LPT, 3, 1)){
#else
        if(ioperm(LPT, 3, 1)) {
#endif
            perror("ioperm(1)");
            return -1;
        }
#else
#ifdef NO_PORT_TALK
        if(lptinit() < 0)
            return -1;
#else
        OpenPortTalk();
#endif
#endif
	out_b(0, LPT);
	fprintf(stderr, "serp> ");
	while(run) {
		if(scanf("%c", &c) < 1) break;
		switch(c) {
			case 'i':
			case 'I':
				main_24xx();
				break;
			case 'm':
			case 'M':
				main_93xx();
				break;
			case 'q':
			case 'Q':
				run = 0;
				break;
			case '\n':
				st = 1;
				break;
			case 'h':
			case 'H':
			case '?':
				usage();
				break;
		}
		if(run && st)
		{
		    fprintf(stderr, "serp> ");
		    st = 0;
		}
	}
#if !defined(_WINDOWS) || defined(_IOPERM)
#ifdef __FreeBSD__
        if(i386_set_ioperm(LPT, 3, 0)){
#else
        if(ioperm(LPT, 3, 0)) {
#endif
            perror("ioperm(1)");
            return -1;
        }
#else
#ifdef NO_PORT_TALK
        lptfree();
#else
        ClosePortTalk();
#endif
#endif
	fprintf(stderr, "\n");
	return 0;
}
